/*
 Copyright 2002, 2003 Alexandre Gil
                                                                                
 This file is part of mwu-extractor.
                                                                                
 mwu-extractor is free software; you can redistribute it and/or modify
 it under the terms of the GNU General Public License as published by
 the Free Software Foundation; either version 2 of the License, or
 (at your option) any later version.
                                                                                
 mwu-extractor is distributed in the hope that it will be useful,
 but WITHOUT ANY WARRANTY; without even the implied warranty of
 MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 GNU General Public License for more details.
                                                                                
 You should have received a copy of the GNU General Public License
 along with mwu-extractor; if not, write to the Free Software
 Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
*/

// Nome: Matrix.cpp
// Objectivo: Implementacao da classe Matrix.
// Autor: Alexandre Gil
// Data: 08 Fev 2002

#include "Matrix.h"
#include "Corpus.h"
#include "NGramMask.h"
#include "NGram.h"
#include "ISVector.h"
#include "MyUtils.h"
#include "Globals.h"

#include <iomanip>

Matrix::Matrix(const Corpus& c, const NGramMask& masks):
	c(c),
	masks(masks),
  dummy(new MNode)
{
	nMasks = masks.NumOfMasks();
	cDim = c.Dim();
	nFiles = c.NumOfFiles();
	
	value = new Matrix::PVNodeVector[nMasks*nFiles];

	dummy->GrowTo(1);
	dummy->Add(Corpus::Pos(0,0));

	InitMatrix();

	TimeStamp(); std::cout << " Vou calcular MEs" << std::endl;
	InitMEValues();

	lastMask = -1;			// basta para obrigar a procurar
}

Matrix::~Matrix()
{
	::PMNode a;
  unsigned long freq, f;
	unsigned long i;
	unsigned short y;

	for(short int x = 1; x < nMasks; x++) 
	{
		for(y = 0; y < nFiles; y++)    // Apagar MNodes
		{
			Matrix::PVVNode pvn = value[x + y*nMasks].pVVNode;
			if(pvn != 0)
			{
				for(i = 0; i < pvn->Filled(); i++)
		  	{
		  		a = pvn->At(i).GetPMNode();
		  		if(a != 0)
		  		{
		  			freq = a->Freq();
		  			for(f = 0; f < freq; f++)
							CleanMNode(a->Pos(f), x);
						delete a;
		  		}
		  	}
			}
		}
		for(y = 0; y < nFiles; y++)  // Apagar os Vector<VNode>
		{
			Matrix::PVVNode pvn = value[x + y*nMasks].pVVNode;
			if(pvn != 0)
 			  delete pvn;
		}
	}
	for(y = 0; y < nFiles; y++)  // Apagar os Vector<VNodeZero>
		{
			Matrix::PVVNodeZero pvn = value[y*nMasks].pVVNodeZero;
			if(pvn != 0)
 			  delete pvn;
		}

	delete dummy;
	delete [] value;
}

void Matrix::CleanMNode(const Corpus::Pos& x, const short int y)
{
	unsigned long p1 = x.FilePos();
	unsigned short f1 = x.FileNum();
	
	Matrix::PVVNode pvn = value[y + f1 * nMasks].pVVNode;
	unsigned long i = pvn->IndexOfBin(VNode(p1));
	pvn->At(i).ResetPMNode();
}

std::ostream& operator<<(std::ostream& output, Matrix& p)
{
	p.Write(output,false);
	return output;
}

void Matrix::Write(std::ostream& output, const bool onlyCount)
{
	unsigned long nulos = 0;
	unsigned long dummys = 0;
	unsigned long filled = 0;
	unsigned long filledZero = 0;

	for(Corpus::Pos x = c.FirstPos(); x <= c.LastPos(); x.Next(c))      // mask 0
		if(FreqAt(x,0) == 0)
		{
			if(!onlyCount) output << "Matrix[" << x << "," << 0 << "]= NULL" << std::endl;
			nulos++;
		}
		else if(FreqAt(x,0) == 1)
		{
			if(!onlyCount) output << "Matrix[" << x << "," << 0 << "]= DUMMY" << std::endl;
			dummys++;
		}
		else
		{
			if(!onlyCount) output << "Matrix[" << x << "," << 0 << "]= " << FreqAt(x,0) << std::endl;
			filledZero++;
		}
	
	for(short int  y = 1; y < nMasks; y++)
		for(Corpus::Pos x = c.FirstPos(); x <= c.LastPos(); x.Next(c))
			if(At(x,y) == 0)
			{
				if(!onlyCount) output << "Matrix[" << x << "," << y << "]= NULL" << std::endl;
				nulos++;
			}
			else if(At(x,y) == dummy)
			{
				if(!onlyCount) output << "Matrix[" << x << "," << y << "]= DUMMY" << std::endl;
				dummys++;
			}
			else
			{
				if(!onlyCount) output << "Matrix[" << x << "," << y << "]= " << (*(At(x,y))) << std::endl;
				filled++;
			}

	output << "Dummys= " << dummys << " Nulos= " << nulos << " filled= " << filled << " filledZero= " << filledZero << " TOTAL " << (filled+filledZero+nulos+dummys) << std::endl;
}

::PMNode Matrix::At(const Corpus::Pos& x, const short int y)
{
	unsigned long p1 = x.FilePos();
	unsigned short f1 = x.FileNum();

	if(y == lastMask && p1 == lastfPos && f1 == lastFile)
		return lastPMNode;
	else
	{
		if(NGram(x,y).Valid())
		{
			Matrix::PVVNode pvn = value[y + f1 * nMasks].pVVNode;
			if(pvn != 0)
			{
				unsigned long i = pvn->IndexOfBin(VNode(p1));
				::PMNode res;
				if(i == pvn->Filled())
					res = dummy;
				else
					res = pvn->At(i).GetPMNode();
				lastMask = y;
				lastfPos = p1;
				lastFile = f1;
				lastPMNode = res;
				return res;
			}
			else
				return dummy;
		}
		else
			return 0;
	}
}

long Matrix::FreqAt(const Corpus::Pos& x, const short int y)
{
	unsigned long p1 = x.FilePos();
	unsigned short f1 = x.FileNum();

	if(y == lastMask && p1 == lastfPos && f1 == lastFile)
		return lastFreq;
	else
	{
		if(NGram(x,y).Valid())
		{
			Matrix::PVVNodeZero pvn = value[y + f1 * nMasks].pVVNodeZero;
			if(pvn != 0)
			{
				unsigned long i = pvn->IndexOfBin(VNodeZero(p1));
				long res;
				if(i == pvn->Filled())
					res = 1;
				else
					res = pvn->At(i).Freq();
				lastMask = y;
				lastfPos = p1;
				lastFile = f1;
				lastFreq = res;
				return res;
			}
			else
				return 1;
		}
		else
			return 0;
	}
}

void Matrix::InitMatrix()
{
  NGram::SetBase(&c,&masks);
  MNode::SetCorpus(&c);

	Corpus::Pos p1;
	short int z;

	Token actual;
	Byte nTokens;
  unsigned long i, j, x;
	NGram ngAnte, ngActual;
  Token previous;
  unsigned long ini;

 	ISVector<NGram> ngv(cDim);

	Corpus::Pos p = c.FirstPos();								// preencher com todas as posicoes do Corpus
	for(j = 0; j < cDim; j++)
	{
		ngv.PutAt(j, new NGram(p, 0));
		p.Next(c);
	}

	TimeStamp(); std::cout << " Ordenar ngv zero" << std::endl;
	ngv.Sort(0, cDim-1);													// ordenar Corpus
	TimeStamp(); std::cout << " Ordenado ngv zero. cDim=" << cDim << " ngv.Filled()=" << ngv.Filled() << std::endl;

	i = 0;																	// ponto de insercao
	j = 0;																	// contador
	ini = 0;																// inicio de cada ciclo

	NGram::SetFirst(0);
	while(j<ngv.Filled())														// Eliminar Tokens com Freq=1
	{
		do
			j++;
		while( (j < ngv.Filled()) && (*(ngv.At(j)) == *(ngv.At(ini))));
		if((j - ini) == 1)
			delete ngv.At(ini);
		else if(ini != i)
			for(x = ini; x < j; x++)
				ngv.PutAt(i++,ngv.At(x));
		else
			i = j;
		ini = j;
	}
	ngv.Reset(i);
	TimeStamp(); std::cout << " Comprimido ngv zero. Total Tokens com freq > 1= " << ngv.Filled() << std::endl;
	
	totalNGrams = 0;
	totalVNodes = 0;
	totalVNodesZero = 0;
	totalMNodes = 0;
	invalid = 0;
	
	FillMatrix(0, ngv);

	for(z = 1; z < nMasks; z++)
	{
  	ini = 0;
	  previous = c.At(ngv.At(0)->Pos());
		nTokens = masks.At(z)->NumOfTokens();
		for(j = 1; j < ngv.Filled(); j++)
		{
			ngv.At(j)->IncMask(); 										 // Actualiza Mask
			ngv.At(j)->SetR(0);
			actual = c.At(ngv.At(j)->Pos());
			if (actual != previous)
			{
				if((j - ini) > 1)
					ngv.Sort(ini, j-1, 1, nTokens-1);
				ini = j;
			}
			previous = actual;
		}
		if((j - ini) > 1)
			ngv.Sort(ini, j-1, 1, nTokens-1);
		FillMatrix(z, ngv);
	}

	for(j = 0; j < ngv.Filled(); j++)         // Apagar ISVector<NGram>
		delete ngv.At(j);

	TimeStamp(); std::cout << " Total n-gramas = " <<  totalNGrams << std::endl;
	TimeStamp(); std::cout << " Total n-gramas com freq > 1 (VNodes) = " << totalVNodes << std::endl;
	TimeStamp(); std::cout << " Total n-gramas zero com freq > 1 (VNodesZero) = " << totalVNodesZero << std::endl;
	TimeStamp(); std::cout << " Total n-gramas diferentes com freq > 1 (MNodes) = " << totalMNodes << std::endl;
	TimeStamp(); std::cout << " Total de invalidos = " << invalid << std::endl;
}

void Matrix::FillMatrix(const short int mask, const ISVector<NGram>& ngv)
{
	unsigned long numberOf = ngv.Filled();
	unsigned long r;                       // Numero de repeticoes em cada loop
	unsigned long j;
	unsigned long i;
	unsigned long * difs1 = new unsigned long[nFiles];
	
	unsigned long k;											 // indice do inicio do ciclo
	::PNGram p;                   				 // Pointer para o NGram no inicio dum ciclo
	::PMNode pNode;												 // Pointer para novo MNode

	if(numberOf > 0)
	{
		for(i = 0; i < nFiles; i++)
			difs1[i] = 0;

		p = ngv.At(0);
		r = 1;
		k = 0;

		NGram::SetFirst(0);
	  for(i = 1; i <= numberOf; i++)				// contar
	  {
	  	if(i < numberOf && (*(ngv.At(i)) == *(ngv.At(i-1))))
	  		r++;
	  	else
	  	{
				while(r > 1 && k < i)
				{
					if(ngv.At(k)->Valid())
						difs1[ngv.At(k)->Pos().FileNum()]++;
					k++;
				}
				p->SetR(r);
  			totalNGrams += r;
				if(i < numberOf)
				{
					p = ngv.At(i);
					k = i;
					r = 1;
				}
	  	}
		}

		for(i = 0; i < nFiles; i++)
		{
			if(difs1[i] > 0)
			{
				if(mask == 0)
				{
					value[mask + i*nMasks].pVVNodeZero = new Vector<VNodeZero>(difs1[i]);
					totalVNodesZero += difs1[i];
				}
				else
				{
					value[mask + i*nMasks].pVVNode = new Vector<VNode>(difs1[i]);
					totalVNodes += difs1[i];
				}
			}
			else
			{
				if(mask == 0)
					value[mask + i*nMasks].pVVNodeZero = 0;
				else
					value[mask + i*nMasks].pVVNode = 0;
			}
		}

		i = 0;
	  while(i < numberOf)					// inserir elementos
	  {
	  	r = ngv.At(i)->GetR();
			if(ngv.At(i)->Valid())
	  	{
		  	if(r > 1)
				{
					if(mask == 0)
					{
		  			for(j = 0; j < r; j++)
		  			{
		  				p = ngv.At(i++);
							value[mask + p->Pos().FileNum()*nMasks].pVVNodeZero->Add(VNodeZero(p->Pos().FilePos(), r));
			 			}
					}
					else
					{
						totalMNodes++;
		 				pNode = new MNode();
		 				pNode->GrowTo(r);
		  			for(j = 0; j < r; j++)
		  			{
		  				p = ngv.At(i++);
							value[mask + p->Pos().FileNum()*nMasks].pVVNode->Add(VNode(p->Pos().FilePos(), pNode));
			 				pNode->Add(p->Pos());
			 			}
					}
	 			}
	 			else
	 				i++;
			}
			else
			{
				i += r;
				invalid += r;
			}
		}

		for(i = 0; i < nFiles; i++)
			if(difs1[i] > 0)
			{
				if(mask == 0)
					value[mask + i*nMasks].pVVNodeZero->Sort(0, difs1[i]-1);
				else
					value[mask + i*nMasks].pVVNode->Sort(0, difs1[i]-1);
			}
	}
	delete [] difs1;
}

const ::PMNode Matrix::Dummy() const
{
	return dummy;
}

void Matrix::InitMEValues()
{
	::PMNode node;
	unsigned long i;
	unsigned short y;

	ResetAllNodes();
	
	for(short int x = 1; x < nMasks; x++) 
	{
		for(y = 0; y < nFiles; y++) 
		{
			PVVNode pvn = value[x + y*nMasks].pVVNode;
			if(pvn != 0)
				for(i = 0; i < pvn->Filled(); i++)
				{
					node = pvn->At(i).GetPMNode();
					if(!node->IsSet())
					{
						node->SetME(CalcME(Corpus::Pos(y, pvn->At(i).FPos()), x, node->Freq()));
						node->Set();
					}
				}
		}
	}
}

float Matrix::CalcME(const Corpus::Pos& pos, const short int nMask, const unsigned long freq)
{
	float me = 0;
	unsigned long soma;
	Byte nSub;
	short int delta;
	Corpus::Pos pAux;

	if(masks.At(nMask)->Lenght() > 1)
	{
		nSub = masks.At(nMask)->NSub();
		if(nSub > 0)
		{
			soma = 0;
			pAux = pos;
			for(Byte mk = 0; mk < nSub; mk++)
			{
				delta = masks.At(nMask)->SubDelta(mk);
				if(delta != 0)
				{
					pAux.MoveSameFile(delta);
					if(masks.At(nMask)->SubMask(mk) == 0)
						soma += FreqAt(pAux, masks.At(nMask)->SubMask(mk));
					else
						soma += At(pAux, masks.At(nMask)->SubMask(mk))->Freq();
					pAux.MoveSameFile(-delta);
				}
				else
				{
					if(masks.At(nMask)->SubMask(mk) == 0)
						soma += FreqAt(pAux, masks.At(nMask)->SubMask(mk));
					else
						soma += At(pAux, masks.At(nMask)->SubMask(mk))->Freq();
				}
			}
			if(soma > 0)
				me = (static_cast<float>(freq) * static_cast<float>(freq)) / ((static_cast<float>(cDim) / static_cast<float>(nSub)) * static_cast<float>(soma));
		}
	}
	return me;
}

bool Matrix::IsLocalMax(const Corpus::Pos pos, const short int nMask)
{
	PMNode pNode = At(pos, nMask);
	Corpus::Pos pAux;
	Byte mk;
	
	short int delta;
	float maxSub;
	float maxSuper;
	short int superMask;

	unsigned long f, freq;

	bool res = false;

	if(masks.At(nMask)->NumOfGaps() > 0  && masks.At(nMask)->NumOfTokens() >= 2)
	{
		float me = At(pos, nMask)->ME();

		maxSub = 0;
		if(masks.At(nMask)->NumOfTokens() > 2)
		{
			pAux = pos;
			for(mk = 0; mk < masks.At(nMask)->NSub(); mk++)
			{
				delta = masks.At(nMask)->SubDelta(mk);
				if(delta != 0)
				{
					pAux.MoveSameFile(delta);
					maxSub = MAX(maxSub, (masks.At(nMask)->SubMask(mk) == 0 ? 0 :At(pAux, masks.At(nMask)->SubMask(mk))->ME()));
					pAux.MoveSameFile(-delta);
				}
				else
					maxSub = MAX(maxSub, (masks.At(nMask)->SubMask(mk) == 0 ? 0 :At(pAux, masks.At(nMask)->SubMask(mk))->ME()));
			}
		}

		maxSuper = 0;
		freq = pNode->Freq();
		for(f = 0; f < freq; f++)
		{
			pAux = pNode->Pos(f);  
			for(mk = 0; mk < masks.At(nMask)->NSuper(); mk++)
			{
				superMask = masks.At(nMask)->SuperMask(mk);
				delta = masks.At(nMask)->SuperDelta(mk);
				if(superMask != 0)						// mascara zero implica me = 0
				{
					if(delta != 0)
					{
						if(pAux.CanMoveSameFile(c, delta))
						{
							pAux.MoveSameFile(delta);
							if(At(pAux, superMask) != 0)    // testa validade do super-grama 
								maxSuper = MAX(maxSuper, (At(pAux, superMask) == dummy ? CalcME(pAux, superMask,1) : At(pAux, superMask)->ME()));
							pAux.MoveSameFile(-delta);
						}
					}
					else
					{
						if(At(pAux, superMask) != 0)    // testa validade do super-grama 
							maxSuper = MAX(maxSuper, (At(pAux, superMask) == dummy ? CalcME(pAux, superMask,1) : At(pAux, superMask)->ME()));
					}
				}
			}
		}
		res = (maxSub <= me) && (me > maxSuper);
	}
	return res;
}

void Matrix::ResetAllNodes()
{
	unsigned long i;
	unsigned short y;

	for(short int x = 1; x < nMasks; x++) 
		for(y = 0; y < nFiles; y++) 
		{
			PVVNode pvn = value[x + y*nMasks].pVVNode;
			if(pvn != 0)
				for(i = 0; i < pvn->Filled(); i++)	
		  		pvn->At(i).GetPMNode()->Reset();
		}
}

void Matrix::WriteMWUs(std::ofstream& fout, const bool restrict)
{
	PMNode pNode;
	float me;
	Byte m;
	long freq;

	float gerados = 0;
	
	unsigned long i;
	unsigned short x;
	short int y;
	Corpus::Pos p;

	TimeStamp(); std::cout << " Inicio desmarcar nodes " << std::endl;
	ResetAllNodes();
	TimeStamp(); std::cout << " Inicio Write MWus " << std::endl;
	
	for(y = 1; y < nMasks; y++)
	{
		std::cout << "                                  \r";
		std::cout << gerados << "\t" << y << "\r";
		if(masks.At(y)->NumOfGaps() > 0) 		// && (!restrict || restrict && masks.At(y)->Valid()))
		{
			for(x = 0; x < nFiles; x++) 
			{
				PVVNode pvn = value[y + x*nMasks].pVVNode;
				if(pvn != 0)
				{
					for(i = 0; i < pvn->Filled(); i++)
					{
			  		pNode = pvn->At(i).GetPMNode();
		  			if(!pNode->IsSet()) 
		  			{
		  				p = Corpus::Pos(x, pvn->At(i).FPos());
	  					if(IsLocalMax(p, y))
							{
								m = masks.At(y)->Lenght();
								fout.write((char *)&m, sizeof(Byte));					
			  				me = pNode->ME();
								fout.write((char *)&me, sizeof(float));
								freq = pNode->Freq();
								fout.write((char *)&freq, sizeof(long));
								WriteNGram(fout, p, y);
	//  					std::cout << "  " <<std::setw(16) << std::setprecision(14) << std::setfill('0') << std::setiosflags(std::ios::fixed) << me;
	//  					std::cout << " Freq= " << freq << " m= " << (int)m << std::endl;
								gerados++;
							}
							pNode->Set();
		  			}
					}
				}
			}
	  }
	}
	std::cout << "                                                         \r";
	std::cout << "Total MWUs= " << gerados << std::endl;
}

void Matrix::WriteNGram(std::ostream& output, const Corpus::Pos pos, const short int nMask) const
{
	Token tk;
	short int k = masks.At(nMask)->Lenght();
	Corpus::Pos p = pos;
	for(int i = 0; i < k; i++)
	{
		if(masks.At(nMask)->At(i) == GAP)
			tk = 0;
		else
			tk = c.At(p);
		output.write((char *)&tk, sizeof(Token));
		p.Next(c);
	}
}


