/*
 Copyright 2002, 2003 Alexandre Gil
                                                                                
 This file is part of mwu-extractor.
                                                                                
 mwu-extractor is free software; you can redistribute it and/or modify
 it under the terms of the GNU General Public License as published by
 the Free Software Foundation; either version 2 of the License, or
 (at your option) any later version.
                                                                                
 mwu-extractor is distributed in the hope that it will be useful,
 but WITHOUT ANY WARRANTY; without even the implied warranty of
 MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 GNU General Public License for more details.
                                                                                
 You should have received a copy of the GNU General Public License
 along with mwu-extractor; if not, write to the Free Software
 Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
*/

// Nome: NGramMask.cpp
// Objectivo: Implementacao da classe generica NGramMask.
// Autor: Alexandre Gil
// Data: 09 Dez 2001

#include "NGramMask.h"
#include "Mask.h"
#include "Globals.h"
#include "ISVector.h"

NGramMask::NGramMask(const short int m, const bool onlyValid):
  ISVector<Mask>(2),
  m(m),
  onlyValid(onlyValid)
{
	Mask ng(m);

	for(short int i=0; i<m; i++)
	  ng[i] = GAP;

	GenNGram(ng,0,0);

	InitSubSets();
	InitSuperSets();
}

NGramMask::~NGramMask()
{
	for(unsigned i=0; i<Filled(); i++)
		delete At(i);
}

void NGramMask::GenNGram(const Mask& mk, const int level, const int depth)
{
	if(level < m)
	{
	  Mask mk1(m);
		for(int i = 0; i < m; i++)
		  mk1[i] = mk[i];

		if(depth == 0)
		{
			mk1[level] = FILLED;
			if(!onlyValid || mk1.Lenght() == 1 || (onlyValid && mk1.Valid()))
				PutAt(Filled(), new Mask(mk1)); // Acrescentar a Mask ao vector de Masks	
		}
		if((!onlyValid && depth < (m - 2)) || (onlyValid && depth < ((m - 1)/2 -1)))  // max #GAPS
			GenNGram(mk1,level+1,depth+1);      // tenta mais  esquerda com GAPS
	  GenNGram(mk1,level+1,0);							// vai para a direita com FILLED
	}	
}

short int NGramMask::NumOfMasks() const
{
	return Filled();
}

short int NGramMask::LeftFilledMask(short int nLeft) const
{
	Mask mk(m);
	for(short int i = 0; i < nLeft; i++)
		mk[i] = FILLED;

	return IndexOf(mk);
}

void NGramMask::InitSubSets()
{
	Mask aux0(m);
	Mask aux(m);

	Byte l, j;
	short int nMask, delta;

	for(int i = 0; i < NumOfMasks(); i++)
	{
		l = At(i)->Lenght();
		if(l > 1)
		{
			At(i)->IniSub();
			aux0 = *At(i);
			aux = *At(i);
			delta = aux0.ShiftLeft();
			nMask = IndexOf(aux0);
			if(nMask != (short int)Filled())
				At(i)->AddSub(nMask, delta);     // Add para o Token na posicao zero
			for(j = delta;j < l;j++)
				if(aux[j] != GAP)
				{
					aux[j] = GAP;
					nMask = IndexOf(aux);
					if(nMask != (short int)Filled())
						At(i)->AddSub(nMask, 0);
					aux[j] = FILLED;
				}
		}
	}
}

void NGramMask::InitSuperSets()
{
	Mask aux(m);

	Byte j;

	for(int i = 0; i < NumOfMasks(); i++)
	{
		short int nMask, delta;

		if(At(i)->Lenght() > 1)
		{
			At(i)->IniSuper();

			// substituir GAPS e adicionar Tokens para a direita
			aux = *At(i);
			for(j = 1; j < m; j++)
				if(aux[j] == GAP)
				{
					aux[j] = FILLED;
					nMask = IndexOf(aux);
					if(nMask != (short int)Filled())
						At(i)->AddSuper(nMask, 0);
					aux[j] = GAP;				
				}

			// adicionar Tokens para a esquerda
			bool clean = false;
			delta = -1;
			aux = *At(i);
			Byte l = aux.Lenght();
			for(j=0; j < (m - l); j++)
			{
				aux.ShiftRight(clean);
				nMask = IndexOf(aux);
				if(nMask != (short int)Filled())
					At(i)->AddSuper(nMask, delta);
				clean = true;
				delta--;
			}
		}
	}
}
