/*
 Copyright 2002, 2003 Alexandre Gil
                                                                                
 This file is part of mwu-extractor.
                                                                                
 mwu-extractor is free software; you can redistribute it and/or modify
 it under the terms of the GNU General Public License as published by
 the Free Software Foundation; either version 2 of the License, or
 (at your option) any later version.
                                                                                
 mwu-extractor is distributed in the hope that it will be useful,
 but WITHOUT ANY WARRANTY; without even the implied warranty of
 MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 GNU General Public License for more details.
                                                                                
 You should have received a copy of the GNU General Public License
 along with mwu-extractor; if not, write to the Free Software
 Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
*/

// Nome: Corpus.cpp
// Objectivo: Implementacao da abstraccao para ficheiros do Corpus
//            Recebe um corpus (nesta fase um ficheiro binario de n short int=2bytes)
// Autor: Alexandre Gil
// Data: 14 Jan 2002

#include <fstream>
#include <cstring>

#include "MyUtils.h"
#include "Corpus.h"

#include <iostream>
#include <iomanip>
#include <math.h>

// Metodos da subclasse Corpus::Pos ----------------------------------------------------------

Corpus::Pos::Pos(const unsigned short fileNum, const unsigned long pos):
fileNum(fileNum),
pos(pos)
{
}

Corpus::Pos::Pos(const Corpus::Pos& p):
fileNum(p.fileNum),
pos(p.pos)
{
}

Corpus::Pos::Pos():
fileNum(0),
pos(0)
{
}

Corpus::Pos::~Pos()
{
}

void Corpus::Pos::Next(const Corpus& c)
{
	if(pos == (c.file[fileNum].nTokens-1))
	{
		fileNum++;
		pos = 0;
	}
	else
		pos++;
}

void Corpus::Pos::Previous(const Corpus& c)
{
	if(pos == 0)
	{
		fileNum--;
		pos = c.file[fileNum].nTokens-1;
	}
	else
		pos--;
}


bool Corpus::Pos::CanMoveSameFile(const Corpus& c, const int delta) const
{
	return (
	    ( (delta > 0) && (delta < (int)(c.file[fileNum].nTokens - pos)))
	    || ( (delta < 0) && (ABS(delta) <= (int)pos) )
	    || delta == 0
	    );
}

unsigned long Corpus::Pos::MaxMoveSameFile(const Corpus& c) const
{
	return ((c.file[fileNum].nTokens-1) - pos);
}

void Corpus::Pos::MoveSameFile(const int delta)
{
	pos += delta;
}

bool Corpus::Pos::operator<=(const Corpus::Pos &p) const
{
  return((fileNum < p.fileNum) || ((fileNum == p.fileNum) && (pos <= p.pos)));
}

bool Corpus::Pos::operator>=(const Corpus::Pos &p) const
{
  return((fileNum > p.fileNum) || ((fileNum == p.fileNum) && (pos >= p.pos)));
}

bool Corpus::Pos::operator<(const Corpus::Pos &p) const
{
  return((fileNum < p.fileNum) || ((fileNum == p.fileNum) && (pos < p.pos)));
}

bool Corpus::Pos::operator>(const Corpus::Pos &p) const
{
  return((fileNum > p.fileNum) || ((fileNum == p.fileNum) && (pos > p.pos)));
}

bool Corpus::Pos::operator==(const Corpus::Pos &p) const
{
  return((fileNum == p.fileNum) && (pos == p.pos));
}

bool Corpus::Pos::operator!=(const Corpus::Pos &p) const
{
  return((fileNum != p.fileNum) || (pos != p.pos));
}

Corpus::Pos& Corpus::Pos::operator=(const Corpus::Pos& p)
{
	fileNum = p.fileNum;
	pos = p.pos;

	return *this;
}

std::ostream& operator<<(std::ostream& output, const Corpus::Pos& p)
{
	output << std::setw(2) << std::setfill('0') << p.fileNum << "/";
	output << std::setw(5) << std::setfill('0') << p.pos;
	return output;
}

void Corpus::Pos::SetFilePos(unsigned long p)
{
	pos = p;
}

void Corpus::Pos::SetFileNum(unsigned short f)
{
	fileNum = f;
}

unsigned long Corpus::Pos::FilePos() const
{
	return pos;
}

unsigned short Corpus::Pos::FileNum()const
{
	return fileNum;
}


// Metodos da classe Corpus ----------------------------------------------------------

void Corpus::ReadFileName(char * n, int max)
{
	char c;

	int i = 0;
	c = fList.get();
	while(c != '\n' && !fList.eof() && i < max)
	{
		n[i++] = c;
		c = fList.get();
	}
	n[i] = '\0';
	while(c != '\n' && !fList.eof())
	{
		c = fList.get();
	}
}

Corpus::Corpus():
	numFiles(0),
	pos(0,0),
	erro(false)
{
}

Corpus::Corpus(const char * fLName):
	fListName(new char[strlen(fLName)+1]),
	numFiles(0),
	pos(Corpus::Pos(0,0)),
	erro(false)
{
	strcpy(fListName, fLName);

  fList.open(fListName);
	erro = !fList;
	if(!erro)
	{
		char buf[MAXFILENAME];

		ReadFileName(buf, MAXFILENAME);

		while(!erro && !fList.eof() && numFiles<MAXFILES)
		{
			if(strlen(buf) > 0)
			{
				file[numFiles].name = new char[strlen(buf)+1];
				strcpy(file[numFiles].name, buf);
			  file[numFiles].fp.open(file[numFiles].name, std::ios::binary);
				erro = !file[numFiles].fp;
				if(!erro)
				{
					file[numFiles].fp.seekg(0, std::ios::end);
					erro = !file[numFiles].fp;
					if(!erro)
	        {
						file[numFiles].nTokens = file[numFiles].fp.tellg() / sizeof(Token);
						file[numFiles].fp.seekg(0, std::ios::beg);
						erro = !file[numFiles].fp;
						if(!erro)
	        	{
  						file[numFiles].cache = new Token[file[numFiles].nTokens];
	        		ReadCache(numFiles);
	  					if(!erro)
	  					{
								if(numFiles == 0)
									file[numFiles].iBase = 0;
								else
									file[numFiles].iBase = file[numFiles-1].iBase+file[numFiles-1].nTokens;
								file[numFiles].fp.close();
								numFiles++;
								ReadFileName(buf, MAXFILENAME);
		 					}
	        	}
	        }
       	}
			}
			else
				ReadFileName(buf, MAXFILENAME);
		}
		fList.close();
		std::cout << "Lidos " << numFiles << " ficheiros" << std::endl;
	}
}

Corpus::~Corpus()
{
	for(unsigned short i = 0; i<numFiles; i++)
	{
		delete [] file[i].name;
    if(file[i].nTokens > 0)
			delete [] file[i].cache;
	}
  delete [] fListName;
}

bool Corpus::Erro() const
{
	return erro;
}

Corpus::Pos Corpus::LastPos() const
{
	return Pos(numFiles-1,file[numFiles-1].nTokens -1);
}

Corpus::Pos Corpus::FirstPos() const
{
	return Pos(0,0);
}

bool Corpus::DefinedPosition(Corpus::Pos i) const
{
	return(i.FilePos() < file[i.FileNum()].nTokens &&  i>=FirstPos() && i <= LastPos());
}

unsigned short Corpus::NumOfFiles() const
{
	return numFiles;
}

void Corpus::ReadCache(const unsigned short fNum)
{
	Token t;
		
	for(unsigned long p = 0; p < file[fNum].nTokens; p++)
	{
		file[fNum].fp.read((char *)&t, sizeof(Token));
		if(erro = !file[fNum].fp) break;	
		file[fNum].cache[p] = t;
	}
}

Token Corpus::At(Corpus::Pos i) const
{
	return file[i.FileNum()].cache[i.FilePos()];
}


unsigned long Corpus::Dim() const
{
	unsigned long res = 0;
	
	for(unsigned short i = 0; i < numFiles; i++)
		res += file[i].nTokens;

	return res;
}

unsigned long Corpus::Index(const Pos& p) const
{
	return p.FilePos() + file[p.FileNum()].iBase;
}
