/*
 Copyright 2002, 2003 Alexandre Gil
                                                                                
 This file is part of mwu-extractor.
                                                                                
 mwu-extractor is free software; you can redistribute it and/or modify
 it under the terms of the GNU General Public License as published by
 the Free Software Foundation; either version 2 of the License, or
 (at your option) any later version.
                                                                                
 mwu-extractor is distributed in the hope that it will be useful,
 but WITHOUT ANY WARRANTY; without even the implied warranty of
 MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 GNU General Public License for more details.
                                                                                
 You should have received a copy of the GNU General Public License
 along with mwu-extractor; if not, write to the Free Software
 Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
*/

// Nome: ISVector.h
// Objectivo: Interface da classe generica ISVector
//            Inpirado no livro do Prof. Pedro Guerreiro (UNL/FCT)
//            Implenta um contentor de pointers para strings.
// Autor: Alexandre Gil
// Data: 24 Jan 2002

#ifndef _H_ISVector
#define _H_ISVector

#include<iostream>

#define MED3(a,b,c) 														\
   (At(a)->At(depth) < At(b)->At(depth) ?				\
     (At(b)->At(depth) < At(c)->At(depth) ? 		\
				(b)																			\
			: 																				\
				(At(a)->At(depth) < At(c)->At(depth) ?	\
					(c) 																	\
				 :																			\
					(a)																		\
				)																				\
			)                     										\
    : (At(b)->At(depth) > At(c)->At(depth) ? 		\
    		(b)																			\
    	 :																				\
    	 	(At(a)->At(depth) > At(c)->At(depth) ? 	\
    	 		(c)  																	\
         : 																			\
         	(a)																		\
        )																				\
      )																					\
    )																						                                                                                   



template <class T>
class ISVector {
	private:
    typedef T * TPointer;
  private:
		TPointer *items;
		unsigned long capacity;
		double growFactor;
		unsigned long filled;

		bool debug;
	private:
  	void MKQuickSort(unsigned long l, unsigned long r, unsigned short depth, unsigned short maxDepth); // pre upperBound - lowerBound > 0
		unsigned long ChoosePivot(unsigned long l, unsigned long r, unsigned short depth) const;
	public:
		const TPointer At(unsigned long x) const;   	 // pre x >= 0 && x < Capacity()
	  void PutAt(unsigned long x, const TPointer s); // pre x >= 0 && x <= Capacity()
	  void Add(const TPointer s); 
		void Copy(const ISVector<T>& other);

		explicit ISVector(unsigned long capacity);									// pre capacity > 0;
		explicit ISVector(const ISVector<T>& other);
		~ISVector();

		ISVector<T>& operator=(const ISVector<T>& other);

		unsigned long Capacity() const;
		unsigned long Filled() const;

	  double GrowFactor() const;
	  void SetGrowFactor(double newGrowFactor);
	  void Grow();
	  void GrowTo(unsigned long newCapacity);

		bool Has(const T& s) const;								
		unsigned long CountIf(const T& s) const;
		unsigned long IndexOf(const T& s) const;				   
		unsigned long IndexOfBin(const T& s) const;				   
		unsigned long LastIndexOf(const T& s) const;       

    void Swap(unsigned long x, unsigned long y);     							 // pre x >= 0 && x < Capacity()
    																																		   // pre y >= 0 && y < Capacity()
		void Reset(const unsigned long newFilled=0);
		
		void Sort(unsigned long l, unsigned long r, unsigned short depth=0, unsigned short maxDepth=0); // pre r - l > 0

    void Write(std::ostream& output = std::cout) const;

    void SetDebug(bool state);
};

#endif
