/*
 Copyright 2002, 2003 Alexandre Gil
                                                                                
 This file is part of mwu-extractor.
                                                                                
 mwu-extractor is free software; you can redistribute it and/or modify
 it under the terms of the GNU General Public License as published by
 the Free Software Foundation; either version 2 of the License, or
 (at your option) any later version.
                                                                                
 mwu-extractor is distributed in the hope that it will be useful,
 but WITHOUT ANY WARRANTY; without even the implied warranty of
 MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 GNU General Public License for more details.
                                                                                
 You should have received a copy of the GNU General Public License
 along with mwu-extractor; if not, write to the Free Software
 Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
*/

// Nome: Mascara.h
// Objectivo: Interface da classe generica Mask
// Autor: Alexandre Gil
// Data: 01 Fev 2002

#ifndef _H_Mask
#define _H_Mask

#include<iostream>
#include "MyUtils.h"

class Mask
{
	private:
		typedef struct
		{
			short int nMask;
			short int delta;
		} PMask;

		class MRef {
			private:
				Mask &w;
				const Byte index;
			public:
				MRef(Mask &w, Byte i): w(w), index(i) {}
				// Para escrita
				MRef &operator=(Byte s) {	w.PutAt(s,index);	return *this;	}
				MRef &operator=(const MRef &r) { return *this = (Byte)(r); }
				operator Byte() const {return w.At(index); }
		};
		friend class Mask::MRef;

	private:
		Byte * value;
		Byte * pFilled;
		
		Byte m;
		bool valid;
		Byte lenght;
		Byte nTokens;

		PMask * pSub;
		PMask * pSuper;

		Byte nSub;
		Byte nSuper;

		static Byte first;

	private:
	  void PutAt(const Byte v, Byte x);
	  void Copy(const Mask& other);
	  void Delete();
	  void SetValidity();

	public:
		Mask();
		Mask(Byte m);
		Mask(const Mask& other);
		~Mask();

		static void SetFirst(const Byte newFirst);
		Byte At(const Byte x) const;      

		MRef operator[](Byte i);   				             // pre i >= 0 && i < MDim()
		Byte operator[](Byte i) const { return At(i);}; // pre i >= 0 && i < MDim()
		Mask& operator=(const Mask& other);

		Byte MDim() const;
		Byte Lenght() const;
		Byte NumOfTokens() const;
		Byte NumOfGaps() const;
		bool Valid() const;
		
		friend std::ostream& operator<<(std::ostream& output, const Mask& p);

		bool operator==(const Mask& s) const;
		bool operator>(const Mask& s) const;
		bool operator<(const Mask& s) const;
		bool operator>=(const Mask& s) const;
		bool operator<=(const Mask& s) const;

		void IniSub();
		void IniSuper();

		Byte MaxSub() const;
		Byte MaxSuper() const;

		Byte NSub() const;
		Byte NSuper() const;

		void AddSub(short int nMask, short int delta);
		void AddSuper(short int nMask, short int delta);

		Byte ShiftLeft();      			// pre nTokens > 1; devolve numero de casas rodadas
																				// roda para a esquerda de tal modo que o segundo 
																				// token passe para a posicao 0
		void ShiftRight(bool clean);// roda uma posicao para a direita e coloca
																				// posicao 0 a FILLED. Coloca anterior posicao 0
																				// com GAP se clean = TRUE

		short int SubMask(Byte x) const;				//pre x >= 0 && x < nSub
		short int SubDelta(Byte x) const;			//pre x >= 0 && x < nSub

		short int SuperMask(Byte x) const;			//pre x >= 0 && x < nSuper
		short int SuperDelta(Byte x) const;			//pre x >= 0 && x < nSuper

		Byte Index(Byte i) const;                //pre i>0 && i < nTokens
};



#endif

