/*
 Copyright 2002, 2003 Alexandre Gil
                                                                                
 This file is part of mwu-extractor.
                                                                                
 mwu-extractor is free software; you can redistribute it and/or modify
 it under the terms of the GNU General Public License as published by
 the Free Software Foundation; either version 2 of the License, or
 (at your option) any later version.
                                                                                
 mwu-extractor is distributed in the hope that it will be useful,
 but WITHOUT ANY WARRANTY; without even the implied warranty of
 MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 GNU General Public License for more details.
                                                                                
 You should have received a copy of the GNU General Public License
 along with mwu-extractor; if not, write to the Free Software
 Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
*/

// Nome: NGram.cpp
// Objectivo: Implementacao da Classe NGram. 
// Autor: Alexandre Gil
// Data: 02 Fev 2002


#include "NGram.h"

#include "Corpus.h"
#include "Globals.h"
#include "MyUtils.h"

#include <iostream>
#include <iomanip>

Corpus const * MNode::pCorpus;
Corpus const * NGram::pCorpus;
NGramMask const * NGram::pvMask;
Byte NGram::first;


NGram::NGram():
	pos(Corpus::Pos(0,0)),
	pMask(0)
{
}

NGram::NGram(const Corpus::Pos& pos, const short int pMask):
	pos(pos),
	pMask(pMask)
{
}

NGram::NGram(const NGram& other):
	pos(other.pos),
	pMask(other.pMask)
{
}

NGram::~NGram()
{
}

NGram& NGram::operator=(const NGram& other)
{
	pMask = other.pMask;
	pos = other.pos;

	return *this;
}

void NGram::SetBase(const Corpus * c, const NGramMask * vMask)
{
	pCorpus = c;
	pvMask = vMask;
}

void NGram::SetFirst(const Byte firstToken)
{
	first = firstToken;
}

bool NGram::Valid() const
{
	return pos.CanMoveSameFile(*pCorpus, pvMask->At(pMask)->Lenght()-1);
}

void NGram::IncMask()
{
	pMask++;
}

bool NGram::operator==(const NGram& other) const
{
	if(pMask == other.pMask && Valid() && other.Valid())
	{	
		Corpus::Pos p1 = pos;
		Corpus::Pos p2 = other.pos;

		Byte t = pvMask->At(pMask)->NumOfTokens();
		Byte j = 0;
		Byte w;
		for(Byte i = first; i < t; i++)
		{
			w = pvMask->At(pMask)->Index(i) -j;
			j += w;
			p1.MoveSameFile(w);
			p2.MoveSameFile(w);
			if(pCorpus->At(p1) != pCorpus->At(p2))
				return false;
		}
		return true;	
	}
	else
		return !Valid() && !other.Valid();
}

bool NGram::operator>(const NGram& other) const
{
	return GreaterThan(other, true);
}

bool NGram::operator>=(const NGram& other) const
{
	return GreaterThan(other, false);
}

bool NGram::GreaterThan(const NGram& other, bool restrict) const
{
	if(Valid() && other.Valid())
	{
		Corpus::Pos p1 = pos;
		Corpus::Pos p2 = other.pos;

		if(pMask == other.pMask)
		{

			Byte t = pvMask->At(pMask)->NumOfTokens();
			Byte j = 0;
			Byte w;
			for(Byte i = first; i < t; i++)
			{
				w = pvMask->At(pMask)->Index(i) -j;
				j += w;
				p1.MoveSameFile(w);
				p2.MoveSameFile(w);
				if(pCorpus->At(p1) != pCorpus->At(p2))
					return pCorpus->At(p1) > pCorpus->At(p2);
			}
			return !restrict;	
		}
		else
		{
			short int k1 = pvMask->At(pMask)->Lenght();
			short int k2 = pvMask->At(other.pMask)->Lenght();
			for(int i = 0; i < MIN(k1,k2); i++)
			{
				if(pvMask->At(pMask)->At(i) != GAP && pvMask->At(other.pMask)->At(i) != GAP && pCorpus->At(p1) != pCorpus->At(p2))
					return pCorpus->At(p1) > pCorpus->At(p2);
				else if(pvMask->At(pMask)->At(i) != GAP && pvMask->At(other.pMask)->At(i) == GAP)
					return false;
				else if(pvMask->At(pMask)->At(i) == GAP && pvMask->At(other.pMask)->At(i) != GAP)
					return true;
				p1.Next(*pCorpus);
				p2.Next(*pCorpus);
			}
			return(k1 > k2);
		}
	}
	else
		return Valid();
}

bool NGram::operator<(const NGram& other) const
{
	return LessThan(other, true);
}

bool NGram::operator<=(const NGram& other) const
{
	return LessThan(other, false);
}

bool NGram::LessThan(const NGram& other, bool restrict) const
{
	if(Valid() && other.Valid())
	{
		Corpus::Pos p1 = pos;
		Corpus::Pos p2 = other.pos;

		if(pMask == other.pMask)
		{
			Byte t = pvMask->At(pMask)->NumOfTokens();
			Byte j = 0;
			Byte w;
			for(Byte i = first; i < t; i++)
			{
				w = pvMask->At(pMask)->Index(i) -j;
				j += w;
				p1.MoveSameFile(w);
				p2.MoveSameFile(w);
				if(pCorpus->At(p1) != pCorpus->At(p2))
					return pCorpus->At(p1) < pCorpus->At(p2);
			}
			return !restrict;	
		}
		else
		{
			short int k1 = pvMask->At(pMask)->Lenght();
			short int k2 = pvMask->At(other.pMask)->Lenght();

			for(int i = 0; i < MIN(k1,k2); i++)
			{
				if(pvMask->At(pMask)->At(i) != GAP && pvMask->At(other.pMask)->At(i) != GAP && pCorpus->At(p1) != pCorpus->At(p2))
					return pCorpus->At(p1) < pCorpus->At(p2);
				else if(pvMask->At(pMask)->At(i) != GAP && pvMask->At(other.pMask)->At(i) == GAP)
					return true;
				else if(pvMask->At(pMask)->At(i) == GAP && pvMask->At(other.pMask)->At(i) != GAP)
					return false;
				p1.Next(*pCorpus);
				p2.Next(*pCorpus);
			}
			return(k1 < k2);
		}
	}
	else
		return !Valid();
}

std::ostream& operator<<(std::ostream& output, const NGram& ng)
{
	output << "(" << ng.pos << "/";
	output << std::setw(3) << std::setfill('0') << ng.pMask << ")";

	if(ng.Valid())
	{
		short int k = (ng.pvMask)->At(ng.pMask)->MDim();
		Corpus::Pos p = ng.pos;
		for(int i = 0; i < k; i++)
		{
			if( (ng.pvMask)->At(ng.pMask)->At(i) == GAP )
				output << " <gap>";
			else
				output << " <" << (int)ng.pCorpus->At(p) << ">";
			p.Next(*(ng.pCorpus));
		}
	}
	else
		output << " <INVALID>";
	return output;
}

const Corpus::Pos& NGram::Pos() const
{
	return pos;
}

short int NGram::PMask() const
{
	return pMask;
}

Byte NGram::Lenght() const
{
	return pvMask->At(pMask)->Lenght();
}

Byte NGram::NumOfTokens() const
{
	return pvMask->At(pMask)->NumOfTokens();
}

Byte NGram::NumOfGaps() const
{
	return pvMask->At(pMask)->NumOfGaps();
}

Token NGram::At(const Byte iToken) const
{
	if(Valid())
	{
		Corpus::Pos p1 = pos;
		Byte w = pvMask->At(pMask)->Index(iToken);
		p1.MoveSameFile(w);
		return pCorpus->At(p1);
	}
	return GAP;
}

// *******************************************************************
MNode::MNode():
	freq(0)
{
}

MNode::MNode(const Corpus::Pos& p):
	pos(new Corpus::Pos[1]),
	freq(1)
{
	pos[0] =p;
}

MNode::MNode(const MNode& other):
	freq(other.freq)
{
	if(freq > 0)
	{
		pos = new Corpus::Pos[other.freq];

		for(unsigned long i = 0; i < (unsigned long)freq; i++)
			pos[i] = other.pos[i];
	}
}

MNode::~MNode()
{
	if(freq>0)
		delete [] pos;
}

MNode& MNode::operator=(const MNode& other)
{
	if(freq>0)
		delete [] pos;

	freq = other.freq;
	if(freq>0)
	{
		pos = new Corpus::Pos[freq];
		for(unsigned long i = 0; i < (unsigned long)freq; i++)
			pos[i] = other.pos[i];
	}

	return *this;
}

void MNode::SetCorpus(const Corpus * c)
{
	pCorpus = c;
}


std::ostream& operator<<(std::ostream& output, const MNode& mn)
{
	output << "{" << std::setw(3) << std::setfill('0') << mn.Freq() << "} ";

	for(long i = 0; i < mn.Freq(); i++)
		output << " " << mn.pos[i];

	return output;
}

void MNode::GrowTo(long dim)
{
	if(dim > 0)
	{
		Corpus::Pos* temp = pos;
		pos = new Corpus::Pos[dim];
		if(freq > 0)
		{
			for(unsigned long i = 0; i<(unsigned long)freq && i < (unsigned long)dim; i++)
					pos[i] = temp[i];
			delete [] temp;
		}
	}
	else
	{
		if(freq > 0) delete [] pos;
		freq = 0;
	}
}

void MNode::Add(const Corpus::Pos& p)
{
	pos[freq++] = p;
}

const Corpus::Pos& MNode::Pos(long f) const
{
	return pos[f];
}

long MNode::Freq() const
{
	return ABS(freq);
}

void MNode::Set()
{
	freq = -1*ABS(freq);
}

void MNode::Reset()
{
	freq = ABS(freq);
}

bool MNode::IsSet() const
{
	return freq < 0;
}

