/*
 Copyright 2002, 2003 Alexandre Gil
                                                                                
 This file is part of mwu-extractor.
                                                                                
 mwu-extractor is free software; you can redistribute it and/or modify
 it under the terms of the GNU General Public License as published by
 the Free Software Foundation; either version 2 of the License, or
 (at your option) any later version.
                                                                                
 mwu-extractor is distributed in the hope that it will be useful,
 but WITHOUT ANY WARRANTY; without even the implied warranty of
 MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 GNU General Public License for more details.
                                                                                
 You should have received a copy of the GNU General Public License
 along with mwu-extractor; if not, write to the Free Software
 Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
*/

// Nome: Parser.cpp
// Objectivo: Implementacao da classe Parser
// Autor: Alexandre Gil
// Data: 16 Mar 2002

#include <fstream>
#include <cstring>
#include <ctype.h>

#include "Parser.h"
#include "TSTree.h"
#include "Token.h"

#include <iostream>
#include <iomanip>

void CleanCR(char * s) {
    int i = 0;
    int j = 0;
	
    while (s[i] != '\0') {
	if (s[i] != '\r')
	    s[j++] = s[i];
	i++;
    }	
}

Parser::Parser(TCorpus& c, const char * lexFName, const char * dName):
    erro(false),
    c(c),
    numOfTokens(0)
{
    dirName = new char[strlen(dName)+1];
    strcpy(dirName, dName);
    CleanCR(dirName);

    lexFileName = new char[strlen(lexFName)+1];
    strcpy(lexFileName, lexFName);
    CleanCR(lexFileName);

    std::ifstream f;

    f.open(lexFileName);
    erro = !f;
    if (!erro) {
	dic = new TSTree<Token>(1);
	InitCTab(f);
	f.close();
    }
}

Parser::~Parser() {
    for(int i = 0; i<256; i++)
	if(cTab[i].identifier != 0)
	    delete [] cTab[i].identifier;
    delete dic;
    delete dirName;
    delete lexFileName;
}

int Parser::NextState(const int state, const unsigned char c, const bool eof) {
    int res = -2;   // error by default
    switch(state) {
    case 0:
	if(eof) res = -1;
	else if(isalpha(c)) res = 1;
	else if(c == ' ' || c == '\t' || c == '\n' || c == '\r') res = 0;
	else if(c == '<') res = 8;
	break;
    case 1:
	if(isalpha(c)) res = 1;
	else if(c == ' ' || c == '\t') res = 2;
	else if(c == '=') res = 3;
	else if(c == '|') res = 60;
	break;
    case 2:
	if(c == ' ' || c == '\t') res = 2;
	else if(c == '=') res = 3;
	else if(c == '|') res = 60;
	break;
    case 60:
	if(c == ' ' || c == '\t') res = 60;
	else if(!eof && !isdigit(c) && !isalpha(c)) res = 61;
	break;
    case 61:
	if(c == ' ' || c == '\t') res = 61;
	else if(c == '=' ) res = 3;
	break;
    case 3:
	if(isdigit(c)) res = 4;
	else if(c == ' ' || c == '\t') res = 3;
	break;
    case 4:
	if(isdigit(c)) res = 4;
	else if(c == ' ' || c == '\t' || c == '\r') res = 5;
	else if(c == ',') res = 3;
	else if(c == '\n') res = 0;
	else if(c == '-') res = 15;
	break;
    case 5:
	if(c == ' ' || c == '\t' || c == '\r') res = 5;
	else if(c == ',') res = 3;
	else if(c == '\n') res = 0;
	break;
    case 8:
	if(isalpha(c)) res = 9;
	break;
    case 9:
	if(isalpha(c)) res = 9;
	else if(c == '>') res = 10;
	break;
    case 10:
	if(c == ' ' || c == '\t') res = 11;
	else if(c == '=') res = 12;
	break;
    case 11:
	if(c == ' ' || c == '\t') res = 11;
	else if(c == '=') res = 12;
	break;
    case 12:
	if(isalpha(c)) res = 13;
	else if(c == ' ' || c == '\t') res = 12;
	break;
    case 13:
	if(isalpha(c)) res = 13;
	else if(c == ' ' || c == '\t' || c == '\r') res = 14;
	else if(c == ',') res = 12;
	else if(c == '\n') res = 0;
	break;
    case 14:
	if(c == ' ' || c == '\t' || c == '\r') res = 14;
	else if(c == ',') res = 12;
	else if(c == '\n') res = 0;
	break;
    case 15:
	if(isdigit(c)) res = 16;
	break;
    case 16:
	if(isdigit(c)) res = 16;
	else if(c == ' ' || c == '\t' || c == '\r') res = 5;
	else if(c == ',') res = 3;
	else if(c == '\n') res = 0;
	break;
    default:
	erro = true;
    }
    return res;
}

bool Parser::Has(const char * id, const int from) const
{
    for(int x=from+1; x < 255; x++)
	if(cTab[x].identifier != 0 && strcmp(cTab[x].identifier, id) == 0)
	    return true;
    return false;
}

int Parser::Index(const char * id, const int previous) const
{
    int x;
    for(x=previous+1; x < 255; x++)
	if(cTab[x].identifier != 0 && strcmp(cTab[x].identifier, id) == 0)
	    break;
    return x;
}

void Parser::Action(const int state, const int nextState, const char w)
{
    char aux[2] = " ";
    Token cnt;
    
    switch (state) {
    case -2:
    case -1:
	break;
    case 0:
	if (nextState == 1)
	{
	    pid = 1;
	    identifier[0] = w;
	}
	break;
    case 1:
	if(nextState == 1)
	    identifier[pid++] = w;
	else
	{
	    TokenNum = 0;
	    identifier[pid] = '\0';		
	    erro = Has(identifier);
	}
	break;
    case 2:
	break;
    case 60:
	if(nextState == 61)
	{
	    aux[0] = w;
	    cnt = dic->Count();
	    TokenNum = dic->Insert(aux);
	    erro = (dic->Count() == cnt);    // ja existia
	}
    case 61:
	break;
    case 3:
	if(nextState == 4)
	    value = w - '0';
	break;
    case 4:
	if(nextState == 4)
	    value = value * 10 + w - '0';
	else if(value >= 0 && value < 255)
	{
	    cTab[value].token = TokenNum;
	    cTab[value].isTrash = false;
	    if(cTab[value].identifier == 0)
	    {
		cTab[value].identifier = new char[strlen(identifier)+1];
		strcpy(cTab[value].identifier, identifier);
	    }
	    else
		erro = true;
	}
	else
	    erro = true;
	break;
    case 5:
	break;
    case 8:
	if(nextState == 9)
	{
	    pid = 1;
	    identifier[0] = w;
	}
	break;
    case 9:
	if(nextState == 9)
	    identifier[pid++] = w;
	else
	    identifier[pid] = '\0';
	break;
    case 10:
    case 11:
	break;
    case 12:
	if(nextState == 13)
	{
	    pid = 1;
	    identifier1[0] = w;
	}
	break;
    case 13:
	if(nextState == 13)
	    identifier1[pid++] = w;
	else
	{
	    identifier1[pid] = '\0';
	    if(Has(identifier1))
	    {
		int x = Index(identifier1);
		if(strcmp(identifier, "DOT") == 0)
		{
		    cTab[x].isDot = true;
		    cTab[x].isTrash = false;
		    while(Has(identifier1, x))
		    {
			x = Index(identifier1, x);
			cTab[x].isDot = true;
			cTab[x].isTrash = false;
		    }
		}
		else if(strcmp(identifier, "COMMA") == 0)
		{
		    cTab[x].isComma = true;
		    cTab[x].isTrash = false;
		    while(Has(identifier1, x))
		    {
			x = Index(identifier1, x);
			cTab[x].isComma = true;
			cTab[x].isTrash = false;
		    }
		}
		else if(strcmp(identifier, "HIFFEN") == 0)
		{
		    cTab[x].isHiffen = true;
		    cTab[x].isTrash = false;
		    while(Has(identifier1, x))
		    {
			x = Index(identifier1, x);
			cTab[x].isHiffen = true;
			cTab[x].isTrash = false;
		    }
		}
		else if(strcmp(identifier, "BEGINTRASH") == 0)
		{
		    cTab[x].isBeginTrash = true;
		    cTab[x].isTrash = false;
		    while(Has(identifier1, x))
		    {
			x = Index(identifier1,x);
			cTab[x].isBeginTrash = true;
			cTab[x].isTrash = false;
		    }
		}
		else if(strcmp(identifier, "ENDTRASH") == 0)
		{
		    cTab[x].isEndTrash = true;
		    cTab[x].isTrash = false;
		    while(Has(identifier1, x))
		    {
			x = Index(identifier1,x);
			cTab[x].isEndTrash = true;
			cTab[x].isTrash = false;
		    }
		}
		else if(strcmp(identifier, "LETTERS") == 0)
		{
		    cTab[x].isLetter = true;
		    cTab[x].isTrash = false;
		    while(Has(identifier1, x))
		    {
			x = Index(identifier1,x);
			cTab[x].isLetter = true;
			cTab[x].isTrash = false;
		    }
		}
		else if(strcmp(identifier, "DIGITS") == 0)
		{
		    cTab[x].isDigit = true;
		    cTab[x].isTrash = false;
		    while(Has(identifier1, x))
		    {
			x = Index(identifier1,x);
			cTab[x].isDigit = true;
			cTab[x].isTrash = false;
		    }
		}
		else if(strcmp(identifier, "DELIMITER") == 0)
		{
		    cTab[x].isDelimiter = true;
		    cTab[x].isTrash = false;
		    while(Has(identifier1, x))
		    {
			x = Index(identifier1,x);
			cTab[x].isDelimiter = true;
			cTab[x].isTrash = false;
		    }
		}
	    }	
	    else
		erro = true;
	}
	break;
    case 14:
	break;
    case 15:
	if(nextState == 16)
	    value1 = w -'0';
	break;
    case 16:
	if(nextState == 16)
	    value1 = value1 * 10 + w - '0';
	else if(value <= value1 && value >= 0 && value < 255 && value1 >= 0 && value1 < 255)
	{
	    for(int y = value+1; y <= value1; y++)
	    {
		cTab[y].token = TokenNum;
		cTab[y].isTrash = false;
		if(cTab[y].identifier == 0)
		{
		    cTab[y].identifier = new char[strlen(identifier)+1];
		    strcpy(cTab[y].identifier, identifier);
		    cTab[y].isTrash = false;
		}
		else
		    erro = true;
	    }
	}
	else
	    erro = true;
	break;

    default:
    {
	std::cout << "Error: Default Action" << std::endl;
	erro = true;
    }
    }
}

void Parser::InitCTab(std::ifstream& f)
{
    for(int i=0; i<256; i++)
    {
	cTab[i].isTrash = true;
	cTab[i].isLetter = false;
	cTab[i].isDigit = false;
	cTab[i].isDot = false;
	cTab[i].isComma = false;
	cTab[i].isHiffen = false;
	cTab[i].isDelimiter = false;
	cTab[i].isBeginTrash = false;
	cTab[i].isEndTrash = false;
	cTab[i].token = -1;
	cTab[i].identifier = 0;
    }
    char w;
    int state = 0;
    int nextState;

    while(!erro && state != -1 && state != -2)
    {
	w = f.get();
	nextState = NextState(state, w, f.eof());
	Action(state, nextState, w);
	state = nextState;
    }
    erro = erro || (state == -2);
}

void Parser::WriteCTab() const
{
    for(int i=0; i<256; i++)
    {
	std::cout << i << "\t\t" << cTab[i].isTrash << " ";
	std::cout << cTab[i].isLetter << " ";
	std::cout << cTab[i].isDelimiter << " ";
	std::cout << cTab[i].isBeginTrash << " ";
	std::cout << cTab[i].isEndTrash << " ";
	std::cout << cTab[i].isDigit << " ";
	std::cout << cTab[i].isDot << " ";
	std::cout << cTab[i].isComma << " ";
	std::cout << cTab[i].isHiffen << " ";
	std::cout << cTab[i].token << " ";
	std::cout << (cTab[i].identifier != 0 ? cTab[i].identifier:"Null") << std::endl;
    }	
}

bool Parser::Erro() const
{
    return erro;
}

void Parser::OpenTkFile(std::ofstream& f, std::ofstream& fLstCrp)
{
    char * fileName = new char[strlen(dirName)+strlen(c.FileName())+strlen(TOKEN_FILE)+1];

    strcpy(fileName, dirName);
    strcat(fileName, c.FileName());
    strcat(fileName, TOKEN_FILE);

    f.open(fileName, std::ios::binary);
    erro = !f;
    if(!erro)
	fLstCrp << fileName << std::endl;
    delete fileName;
}

void Parser::WriteTk(std::ofstream& fToken, Token t)
{
    Token x = t;
    fToken.write((char *)&x, sizeof x);
    numOfTokens++;
}

void Parser::GenTokens()
{
    int estado = 0;
    char _word[MAXWORD];
    int wid;
    unsigned char w;
    bool advance = false;
    char a;
    Token tk;

    numOfTokens = 0;

    char * fileName = new char[strlen(dirName)+strlen(c.CorpusName())+strlen(CORPUS_LIST_FILE)+1];

    strcpy(fileName, dirName);
    strcat(fileName, c.CorpusName());
    strcat(fileName, CORPUS_LIST_FILE);

    std::ofstream fLstCrp(fileName);
    erro = !fLstCrp;
    delete fileName;

    std::ofstream fToken;
    if(!erro && !c.AtEndOfCorpus())
	OpenTkFile(fToken, fLstCrp);

    while(!erro && estado != -1 && estado != -2)
    {
	w = c.Current();
	switch(estado)
	{
	case 0:			// estado inicial
	    if (c.AtEndOfCorpus())
		estado = -1;
	    else if (c.AtEOF())
		advance = true;
	    else if(cTab[w].isLetter)
	    {
		wid = 1;
		_word[0] = w;
		estado = 1;
	    }
	    else if(cTab[w].isDigit)
	    {
		wid = 1;
		_word[0] = w;
		estado = 3;
	    }
	    else if(cTab[w].isBeginTrash) estado = 2;
	    else if(cTab[w].isDelimiter || cTab[w].isTrash || cTab[w].isEndTrash) estado = 0;
	    else
		WriteTk(fToken, cTab[w].token);
	    break;
	case 1:			// trata letras
	    if(c.AtEndOfCorpus())
	    {
		_word[wid] = '\0';
		tk = dic->Insert(_word);
		WriteTk(fToken, tk); 
		estado = -1;
	    }
	    else if (c.AtEOF())
	    {
		_word[wid] = '\0';
		tk = dic->Insert(_word);
		WriteTk(fToken, tk); 
		advance = true;
		estado = 0;
	    }
	    else if(cTab[w].isLetter || cTab[w].isHiffen) _word[wid++] = w;
	    else if(cTab[w].isBeginTrash)
	    {
		_word[wid] = '\0';
		tk = dic->Insert(_word);
		WriteTk(fToken, tk); 
		estado = 2;
	    }
	    else if(cTab[w].isDelimiter || cTab[w].isTrash || cTab[w].isEndTrash)
	    {
		_word[wid] = '\0';
		tk = dic->Insert(_word);
		WriteTk(fToken, tk); 
		estado = 0;
	    }
	    else
	    {
		_word[wid] = '\0';
		tk = dic->Insert(_word);
		WriteTk(fToken, tk);
		if(cTab[w].isDigit)
		{
		    wid = 1;
		    _word[0] = w;
		    estado = 3;
		}
		else
		{
		    WriteTk(fToken, cTab[w].token); 
		    estado = 0;
		}
	    }
	    break;
	case 2:					// trata lixo
	    if(c.AtEndOfCorpus()) estado = -1;
	    else if (c.AtEOF())
	    {
		advance = true;
		estado = 0;
	    }
	    else if(cTab[w].isEndTrash) estado = 0;
	    break;
	case 3:				// trata numeros
	    a = -1;
	    if(c.AtEndOfCorpus())
	    {
		if(cTab[(int)_word[wid-1]].isDot || cTab[(int)_word[wid-1]].isComma) a = _word[--wid];
		_word[wid] = '\0';
		tk = dic->Insert(_word);
		WriteTk(fToken, tk); 
		if(a != -1)
		    WriteTk(fToken, cTab[(int)a].token);
		estado = -1;
	    }
	    else if (c.AtEOF())
	    {
		if(cTab[(int)_word[wid-1]].isDot || cTab[(int)_word[wid-1]].isComma) a = _word[--wid];
		_word[wid] = '\0';
		tk = dic->Insert(_word);
		WriteTk(fToken, tk); 
		if(a != -1)
		    WriteTk(fToken, cTab[(int)a].token);
		advance = true;
		estado = 0;
	    }
	    else if(cTab[w].isDigit || cTab[w].isDot || cTab[w].isComma) _word[wid++] = w;
	    else if(cTab[w].isBeginTrash)
	    {
		if(cTab[(int)_word[wid-1]].isDot || cTab[(int)_word[wid-1]].isComma) a = _word[--wid];
		_word[wid] = '\0';
		tk = dic->Insert(_word);
		WriteTk(fToken, tk); 
		estado = 2;
		if(a != -1)
		    WriteTk(fToken, cTab[(int)a].token);
	    }
	    else if(cTab[w].isDelimiter || cTab[w].isTrash || cTab[w].isEndTrash)
	    {
		if(cTab[(int)_word[wid-1]].isDot || cTab[(int)_word[wid-1]].isComma) a = _word[--wid];
		_word[wid] = '\0';
		tk = dic->Insert(_word);
		WriteTk(fToken, tk); 
		if(a != -1)
		    WriteTk(fToken, cTab[(int)a].token); 
		estado = 0;
	    }
	    else
	    {
		if(cTab[(int)_word[wid-1]].isDot || cTab[(int)_word[wid-1]].isComma) a = _word[--wid];
		_word[wid] = '\0';
		tk = dic->Insert(_word);
		WriteTk(fToken, tk); 
		if(a != -1)
		    WriteTk(fToken, cTab[(int)a].token); 
		if(cTab[w].isLetter)
		{
		    wid = 1;
		    _word[0] = w;
		    estado = 1;
		}
		else
		{
		    WriteTk(fToken, cTab[w].token); 
		    estado = 0;
		}
	    }
	    break;
	default:
	{
	    std::cout << "No default" << std::endl;
	    erro = true;
	}
	}

	if(!erro && estado != -1 && estado != -2)
	{
	    if(advance)
	    {
		fToken.close();
		/* erro = !fToken;   ---  mais seguro :-) */
		if(!erro)
		{
		    c.NextFile();
		    if(!erro && !c.AtEndOfCorpus())
			OpenTkFile(fToken, fLstCrp);
		}
		advance = false;
	    }
	    else
		c.NextChar();
	}
    }
    erro = erro || (estado == -2);
    if(!erro)
	fLstCrp.close();
}

void Parser::WriteDic()
{
    char * fileName = new char[strlen(dirName)+strlen(c.CorpusName())+strlen(DIC_FILE)+1];

    strcpy(fileName, dirName);
    strcat(fileName, c.CorpusName());
    strcat(fileName, DIC_FILE);

    std::ofstream fDic(fileName);
    erro = !fDic;
    if(!erro)
    {
	fDic << dic->Count() << std::endl;;
		
	dic->Write(fDic);
	fDic.close();
    }
    delete fileName;
}
	
unsigned long Parser::NumOfTokens() const
{
    return numOfTokens;
}

unsigned long Parser::DicTokens() const
{
    return dic->Count();
}

