/*
 Copyright 2002, 2003 Alexandre Gil
                                                                                
 This file is part of mwu-extractor.
                                                                                
 mwu-extractor is free software; you can redistribute it and/or modify
 it under the terms of the GNU General Public License as published by
 the Free Software Foundation; either version 2 of the License, or
 (at your option) any later version.
                                                                                
 mwu-extractor is distributed in the hope that it will be useful,
 but WITHOUT ANY WARRANTY; without even the implied warranty of
 MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 GNU General Public License for more details.
                                                                                
 You should have received a copy of the GNU General Public License
 along with mwu-extractor; if not, write to the Free Software
 Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
*/

// Nome: Parser.h
// Objectivo: Interface da abstracao para o Parser
//            Recebe um Corpus
// 						Recebe um ficheiro contendo as regras para agrupar os caracteres e a directoria
//            onde esse ficheiro se encontra
//            
// Autor: Alexandre Gil
// Data: 16 Mar 2002


#ifndef _H_Parser
#define _H_Parser

#include <fstream>
#include <iostream>

#include "MyUtils.h"
#include "TCorpus.h"
#include "TSTree.h"
#include "Token.h"

const int MAXIDENTIFIER = 200;
const int MAXWORD = 200;

#define DIC_FILE ".dic"
#define TOKEN_FILE ".int"
#define CORPUS_LIST_FILE ".int"

class Parser
{
	private:
		struct {
			bool isTrash;
			bool isLetter;
			bool isDigit;
			bool isDelimiter;
			bool isBeginTrash;
			bool isEndTrash;
			bool isHiffen;
			bool isDot;
			bool isComma;
			Token token;
			char * identifier;
		} cTab[256];
		bool erro;
		
		TCorpus& c;
		char * dirName;
		char * lexFileName;

		Token TokenNum;

		char identifier[MAXIDENTIFIER];
		char identifier1[MAXIDENTIFIER];
		int pid;
		int value, value1;

	  TSTree<Token> * dic;

	  unsigned long numOfTokens;
		
	private:
		virtual void InitCTab(std::ifstream& f);
		virtual int NextState(const int state, const unsigned char w, const bool eof);
		virtual bool Has(const char * id, const int from = 0) const;
		virtual int Index(const char * id, const int previous = 0) const;
		virtual void Action(const int state, const int nextState, const char w);
		virtual void OpenTkFile(std::ofstream& f, std::ofstream& fLstCrp);
		virtual void WriteTk(std::ofstream& fToken, Token t);
	public:
	  explicit Parser(TCorpus& c, const char * dName, const char * fLName);
	  virtual ~Parser();
	  virtual bool Erro() const;
	  void WriteCTab() const;
		virtual void GenTokens();
		virtual void WriteDic();
		virtual unsigned long NumOfTokens() const;
		virtual unsigned long DicTokens() const;
};


#endif
