/*
 Copyright 2002, 2003 Alexandre Gil
                                                                                
 This file is part of mwu-extractor.
                                                                                
 mwu-extractor is free software; you can redistribute it and/or modify
 it under the terms of the GNU General Public License as published by
 the Free Software Foundation; either version 2 of the License, or
 (at your option) any later version.
                                                                                
 mwu-extractor is distributed in the hope that it will be useful,
 but WITHOUT ANY WARRANTY; without even the implied warranty of
 MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 GNU General Public License for more details.
                                                                                
 You should have received a copy of the GNU General Public License
 along with mwu-extractor; if not, write to the Free Software
 Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
*/

// Nome: Tokenizer.cpp
// Objectivo: Tokenizador. Corpo do programa principal.
//            Recebe um Corpus (conjunto de ficheiros de texto).
//
// Devolve um Corpus na forma duma lista de ficheiros binarios. Cada
// ficheiro contem uma cadeia de Tokens (definido em Token.h).  O
// Corpus deve ter um mximo de ULONG_MAX (4,294,967,295) tokens.
//
// Recebe um ficheiro (full qualified) com a lista de ficheiros (full
// qualified) do Corpus e um ficheiro (full qualified) com as regras
// lexicograficas para extrair os tokens do texto
//
// Devolve (na directoria actual) um ficheiro de texto com a lista de
// ficheiros binarios intermedios, os respectivos ficheiros
// intermedios e um ficheiro de texto separado com o dicionario de
// simbolos.
//
// Os ficheiros intermedios so os originais (sem path) adicionados da
// extensao ".$$$".  O ficheiro do dicionario tem o mesmo nome do
// Corpus (sem path) mas com extensao ".dic".
//
// Autor: Alexandre Gil
// Data: 16 Mar 2002

#include <iostream>
#include <cctype>
#include <iomanip>
#include <fstream>

#include "TCorpus.h"
#include "MyUtils.h"
#include "Parser.h"

int main(int argc, char *argv[])
{
    std::cout << "Tokenizador\n";
	
    if (argc == 3) {
	ClearCRLF(argv[1]);
	TimeStamp();
	std::cout << " Ler ficheiro do Corpus " << argv[1] << std::endl;
	TCorpus c(argv[1]);
	
	if (!c.Erro()) {
	    ClearCRLF(argv[2]);
	    TimeStamp();
	    std::cout << " Ler ficheiro das regras " << argv[2] << std::endl;

	    // podera receber um argumento suplementar para indicar a directoria destino
	    Parser p(c, argv[2], "");			

	    if (!p.Erro()) {
		TimeStamp();
		std::cout << " Vai gerar Tokens." << std::endl;
		p.GenTokens();
		if (!p.Erro()) {
		    TimeStamp();
		    std::cout << " Total Tokens gerados = " << p.NumOfTokens() << std::endl;
		    p.WriteDic();
		    TimeStamp();
		    std::cout << " Ficheiro dicionario gerado com " << p.DicTokens() << " tokens" << std::endl;
		} else{
		    TimeStamp(); std::cout << " Erro na geracao dos Tokens." << std::endl;
		}
	    } else {
		std::cout << "ERRO: Falhou inicializacao do Parser. Ver ficheiro " << argv[2] << std::endl;
		return 1;
	    }
	    return 0;
	} else {
	    std::cout << "ERRO: Falhou inicializacao do Corpus. Ver ficheiro " << argv[1] << std::endl;
	    return 1;
	}
    } else {
	std::cout << "ERRO: Numero errado de argumentos. tk <corpus> <rules>." << std::endl;
	return 1;
    }
}

