/*
 Copyright 2002, 2003 Alexandre Gil
                                                                                
 This file is part of mwu-extractor.
                                                                                
 mwu-extractor is free software; you can redistribute it and/or modify
 it under the terms of the GNU General Public License as published by
 the Free Software Foundation; either version 2 of the License, or
 (at your option) any later version.
                                                                                
 mwu-extractor is distributed in the hope that it will be useful,
 but WITHOUT ANY WARRANTY; without even the implied warranty of
 MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 GNU General Public License for more details.
                                                                                
 You should have received a copy of the GNU General Public License
 along with mwu-extractor; if not, write to the Free Software
 Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
*/

// Nome: Dictionary.cpp
// Objectivo: Implementacao da classe generica Dictionary.
// Autor: Alexandre Gil
// Data: 30 Mar 2001

#include <iostream>
#include <fstream>
#include <iomanip>
#include <ctype.h>
#include <cstring>

#include "Dictionary.h"
#include "MyUtils.h"
#include "Token.h"

Dictionary::Dictionary(const char * filename):
	erro(false)
{
	OpenFile(filename);
	if(!erro)
	{
		numTokens = ReadValue();
		erro = erro || (numTokens <= 0);
		if(!erro)
		{
			char buf[MAXSTRING];
			Token tk;
			
			dVec = new _String[numTokens];
			for(tk = 0; tk < numTokens; tk++)
				dVec[tk] = 0;
				
			while(!erro && !fDic.eof())
			{
				tk = ReadValue(); if(fDic.eof() || erro) break;
				erro = (dVec[tk] != 0); if(erro) break;
				ReadWord(buf);
				if(!erro)
				{
					dVec[tk] = new char[strlen(buf)+1];
					strcpy(dVec[tk], buf);
				}
			}
		}
		fDic.close();
	}
}

Dictionary::~Dictionary()
{
	for(Token i = 0; i < numTokens; i++)
		delete [] dVec[i];
	delete [] dVec;
}

void Dictionary::OpenFile(const char * filename)
{
	fDic.open(filename);
	erro = !fDic;
	if(!erro)
		NextChar();
}

void Dictionary::NextChar()
{
	current = fDic.get();
	erro = !fDic && !fDic.eof();
}

Token Dictionary::ReadValue()
{
	Token t = 0;
	while(!erro && !fDic.eof() && isdigit(current))
	{
		t = t * 10 + (current - '0');
		NextChar();
	}
	
	while(!erro && !fDic.eof() && (current == ' ' || current == '\t' || current == '\n'))
		NextChar();

	return t;
}

void Dictionary::ReadWord(char * s)
{
	int i = 0;
	while(!erro && !fDic.eof() && !(current == ' ' || current == '\t' || current == '\n'))
	{
		s[i++] = current;
		NextChar();
		if(i == MAXSTRING)
		{
			erro = true;
			break;
		}
	}

	if(!erro) s[i] = '\0';
	
	while(!erro && !fDic.eof() && (current == ' ' || current == '\t' || current == '\n'))
		NextChar();	
}

bool Dictionary::Valid(Token tk) const
{
	return tk > 0 && tk < numTokens;	
}

void Dictionary::Write(std::ostream& output) const
{
	output << "Imprimir Dictionary " << std::endl;
	output << "--------------- " << std::endl;
	for(Token i = 0; i < numTokens; i++)
	  output << std::setw(10) << std::setfill('0') << i << " - " <<  (dVec[i]==0? "NULL":dVec[i]) << std::endl;;
	output << "--------------- " << std::endl;
}

bool Dictionary::Error() const
{
	return erro;
}

const char * Dictionary::String(Token tk) const
{
	return dVec[tk];
}
